"""psy-strat psyplot plugin

This module defines the rcParams for the psy-strat plugin. This module will
be imported when psyplot is imported. What is should contain is:

- an rcParams variable as instance of :class:`psyplot.config.rcsetup.RcParams`
  that describes the configuration of your plugin
- a get_versions function that returns the version of your plugin and the ones
  from its requirements

.. warning::

    Because of recursion issues, You have to load the psyplot module before
    loading this module! In other words, you have to type

    .. code-block:: python

        import psyplot
        import psy_strat.plugin"""
from psyplot.config.rcsetup import RcParams
from psy_strat import __version__ as plugin_version
from itertools import repeat
from psy_simple.plugin import (
    rcParams as psys_rcParams, validate_bool, validate_int, try_and_error,
    validate_none, validate_float, validate_str, validate_color,
    validate_lineplot, validate_marker, ValidateList)


def get_versions(requirements=True):
    """Get the versions of psy-strat and it's requirements

    Parameters
    ----------
    requirements: bool
        If True, the requirements are imported and it's versions are included
    """
    ret = {'version': plugin_version}
    if requirements:
        # insert versions of the requirements, e.g. via
        #   >>> import requirement
        #   >>> ret['requirement'] = requirement.__version__
        pass
    return ret


# -----------------------------------------------------------------------------
# ------------------------- validation functions ------------------------------
# -----------------------------------------------------------------------------


# define your validation functions for the values in the rcParams here. If
# a validation fails, the function should raise a ValueError or TypeError


def validate_grouper(value):
    """
    Validate the grouper formatoption

    Parameters
    ----------
    value: tuple (float ``y1``, str ``s``)
        A tuple of length 2, where the first parameter ``0<=y1<=1`` determines
        the distance of the bar to the top y-axis and the second is the title
        of the group
    """
    y1, s = value
    y1 = validate_float(y1)
    s = validate_str(s)
    return y1, s


def validate_axislinestyle(value):
    """Validate a dictionary containing axiscolor definitions

    Parameters
    ----------
    value: dict
        a mapping from 'left', 'right', 'bottom', 'top' to the linestyle

    Returns
    -------
    dict

    Raises
    ------
    ValueError"""
    validate = try_and_error(validate_none, validate_str)
    possible_keys = {'right', 'left', 'top', 'bottom'}
    try:
        value = dict(value)
        false_keys = set(value) - possible_keys
        if false_keys:
            raise ValueError("Wrong keys (%s)!" % (', '.join(false_keys)))
        for key, val in value.items():
            value[key] = validate(val)
    except Exception:
        value = dict(zip(possible_keys, repeat(validate(value))))
    return value


def validate_hlines(value):
    """Validate the hlines formatoption

    Parameters
    ----------
    value: object
        Either None, True or a color"""
    if value is None:
        return value
    elif value is True:
        return '0.9'
    return validate_color(value)


# -----------------------------------------------------------------------------
# ------------------------------ rcParams -------------------------------------
# -----------------------------------------------------------------------------


# define your defaultParams. A mapping from rcParams key to a list of length 3:
#
# 1. the default value
# 2. the validation function of type conversion function
# 3. a short description of the default value
#
# Example::
#
#     defaultParams = {'my.key': [True, bool, 'What my key does']}

psys_validate = psys_rcParams.validate
psys_desc = psys_rcParams.descriptions
defaultParams = {

    # if you define new plotters, we recommend to assign a specific rcParams
    # key for it, e.g.
    'plotter.strat.transpose': [
        True, psys_validate['plotter.simple.transpose'],
        'fmt key to switch x- and y-axis for stratigraphic plots'],
    'plotter.strat.titleprops': [
        {'rotation': 45, 'va': 'bottom', 'ha': 'left',
         'bbox': {'facecolor': 'w', 'edgecolor': 'none'}},
        psys_validate['plotter.baseplotter.titleprops'],
        'The properties of the axes title'],
    'plotter.strat.legendlabels': [
        '%(name)s', psys_validate['plotter.simple.legendlabels'],
        psys_desc['plotter.simple.legendlabels']],
    'plotter.strat.legend': [
        False, psys_validate['plotter.simple.legend'],
        psys_desc['plotter.simple.legend']],
    'plotter.strat.titlesize': [
        'small',
        psys_validate['plotter.baseplotter.titlesize'],
        psys_desc['plotter.baseplotter.titlesize']],
    'plotter.strat.title': [
        '%(name)s',
        psys_validate['plotter.baseplotter.title'],
        'The axes title'],
    'plotter.strat.title_wrap': [
        15, validate_int, 'wrap the title after the given amount of characters'
        ],
    'plotter.strat.hlines': [
        None, try_and_error(validate_none, validate_grouper),
        'Show the measurements'],
    'plotter.strat.grouper': [
        None, try_and_error(validate_none, validate_grouper),
        'Group several plots together using the grouper formatoption'],
    'plotter.strat.grouperweight': [
        psys_rcParams['plotter.baseplotter.titleweight'],
        psys_validate['plotter.baseplotter.titleweight'],
        'Set the fontweight of the grouper text'],
    'plotter.strat.groupersize': [
        psys_rcParams['plotter.baseplotter.titlesize'],
        psys_validate['plotter.baseplotter.titlesize'],
        'Set the fontsize of the grouper text'],
    'plotter.strat.grouperprops': [
        {'bbox': {'facecolor': 'w', 'edgecolor': 'none'}},
        psys_validate['plotter.baseplotter.titleprops'],
        'Set the font properties of the grouper text'],
    'plotter.strat.axislinestyle': [
        None, validate_axislinestyle,
        'The linestyle of the x- and y-axes'],
    'plotter.strat.exag_color': [
        '0.7', psys_validate['plotter.simple.color'],
        'The color for exaggerations'],
    'plotter.strat.exag_factor': [
        10, validate_float, 'The exaggeration factor'],
    'plotter.strat.exag': [
        None, validate_lineplot, 'The plotting style for exaggerations'],
    'plotter.strat.occurences': [
        None, try_and_error(validate_none, validate_float,
                            ValidateList(float, 2)),
        "The range that should be considered as an occurence"],
    'plotter.strat.occurence_marker': [
        '+', validate_marker, 'The symbol of the marker for occurences'],
    'plotter.strat.occurence_value': [
        None, try_and_error(validate_none, validate_float,
                            ValidateList(float)),
        'The value to use for an occurence in the plot']
    }

# create the rcParams and populate them with the defaultParams. For more
# information on this class, see the :class:`psyplot.config.rcsetup.RcParams`
# class
rcParams = RcParams(defaultParams=defaultParams)
rcParams.update_from_defaultParams()



"""Module for a widget for stratigraphic plots

This module defines the :class:`StratPlotsWidget` class that can be used to
manage stratigraphic plots. It is designed as a plugin for the
:class:`psyplot_gui.main.MainWindow` class"""
import sys
from itertools import chain
from psyplot_gui.compat.qtcompat import (
    QWidget, Qt, QTabWidget, QTreeWidget, QTreeWidgetItem, QVBoxLayout,
    QCheckBox, QAbstractItemView)
from psyplot_gui.common import DockMixin
import psyplot.project as psy
from psyplot.utils import unique_everseen


def get_stratplots_widgets(mainwindow=None):
    """Get the :class:`StraditizerWidgets` from the psyplot GUI mainwindow"""
    if mainwindow is None:
        from psyplot_gui.main import mainwindow
    if mainwindow is None:
        raise NotImplementedError(
            "Not running in interactive psyplot GUI!")
    try:
        stratplots = mainwindow.plugins[
            'psy_strat.strat_widget:StratPlotsWidget:stratplots']
    except KeyError:
        raise KeyError('psy_strat not implemented as a GUI plugin!')
    return stratplots


class GrouperItem(QTreeWidgetItem):
    """An item whose contents is filled by a StratGrouper

    This item is automatically initialized by an instance of
    :class:`psy_strat.stratplot.StratGrouper`"""

    def __init__(self, grouper, tree, *args, **kwargs):
        super(QTreeWidgetItem, self).__init__(*args, **kwargs)
        self.grouper = grouper
        self.tree = tree
        self.setText(0, grouper.group)

    def add_array_children(self):
        arrays = self.grouper.arrays
        last = len(arrays) - 1
        group = self.grouper.group
        for i, arr in enumerate(arrays):
            child = QTreeWidgetItem(0)
            self.addChild(child)
            # variable name
            child.setText(0, str(arr.name))
            # arrow buttons
            child.setText(1, u'⇧' if i else '')
            child.setText(2, u'⇩' if i < last else '')
            # checkbox
            cb = QCheckBox()
            cb.setChecked(self.grouper.is_visible(arr))
            cb.stateChanged.connect(self.show_or_hide_func(arr.name))
            self.tree.setItemWidget(child, 3, cb)
            # mean
            child.setText(4, '%1.3f' % arr.mean().values)
            # min
            child.setText(5, '%1.3f' % arr.min().values)
            # max
            child.setText(6, '%1.3f' % arr.max().values)
            # group
            if arr.group != group:
                child.setText(7, str(arr.group))

    def show_or_hide_func(self, name):
        """Create a function that displays or hides the plot for an array"""
        def func(state):
            if state == Qt.Checked:
                self.grouper.show_array(name)
            else:
                self.grouper.hide_array(name)
            self.grouper.figure.canvas.draw()
        return func


class StratPlotsWidget(QWidget, DockMixin):
    """A widget for managing the stratigraphic plots from the psy-strat package
    """

    #: The title of the widget
    title = 'Stratigraphic plots'

    #: Display the dock widget at the right side of the GUI
    dock_position = Qt.RightDockWidgetArea

    @property
    def stratplotter_cls(self):
        """The :class:`psy_strat.plotters.StratPlotter` class if it's
        module has already been imported. Otherwise None."""
        if 'psy_strat.plotters' not in sys.modules:
            return None
        from psy_strat.plotters import StratPlotter, BarStratPlotter
        return (StratPlotter, BarStratPlotter)

    @property
    def hidden(self):
        """True if there is no :class:`psy_strat.plotter.StratPlotter` in the
        current main project"""
        return self.stratplotter_cls is None or not bool(psy.gcp(True)(
            self.stratplotter_cls))

    def __init__(self, *args, **kwargs):
        super(StratPlotsWidget, self).__init__(*args, **kwargs)
        vbox = QVBoxLayout()
        self.tabs = QTabWidget(parent=self)
        self.trees = {}
        self.groupers = {}

        vbox.addWidget(self.tabs)
        self.setLayout(vbox)
        psy.Project.oncpchange.connect(self.update_trees_from_project)

    def update_trees_from_project(self, project):
        """Add a new QTreeWidget from the main project"""
        # do nothing for subprojects and if the StratPlotter has not yet been
        # imported
        if not project.is_main or self.stratplotter_cls is None:
            return
        from psy_strat.stratplot import strat_groupers
        project = project(self.stratplotter_cls)
        datasets = project.datasets
        fnames = project.dsnames_map
        for num in set(self.groupers).difference(datasets):
            del self.groupers[num]
            tree = self.trees.pop(num)
            self.tabs.removeTab(self.tabs.indexOf(tree))
        for num in set(datasets).difference(self.groupers):
            ds = datasets[num]
            ds_arr_names = [
                arr.psy.arr_name for i, arr in enumerate(project)
                if num in project[i:i+1].datasets]
            arrs = project(arr_name=ds_arr_names)
            groupers = []
            for group in unique_everseen(
                    arr.attrs['maingroup'] for arr in arrs):
                arrays = project(maingroup=group)
                identifier = ds[group].identifier
                grouper_cls = strat_groupers[identifier]
                groupers.append(grouper_cls(arrays, use_weakref=True))
            title = 'Dataset %i' % num
            if fnames[num]:
                title += ': ' + fnames[num]
            self.add_tree(groupers, title)
        if self.hidden:
            self.hide_plugin()

    def move_selected_children(self, child, col):
        if col not in [1, 2] or child.parent() is None:
            return
        top = child.parent()
        selected = chain([child], top.tree.selectedItems())
        up = col == 1
        move = -1 if up else 1
        children = sorted(
            unique_everseen(
                [child for child in selected if child.parent() is top],
                key=top.indexOfChild),
            key=top.indexOfChild, reverse=not up)
        current_indices = list(map(top.indexOfChild, children))
        last = top.childCount() - 1
        for current, child in zip(current_indices, children):
            if ((up and current == 0) or (not up and current == last)):
                return
            cb = top.tree.itemWidget(child, 3)
            checked = cb.isChecked()
            top.takeChild(current)
            if up:
                top.insertChild(current + move, child)
            else:
                if current == last - 1:
                    top.addChild(child)
                else:
                    top.insertChild(current + move, child)
            cb = QCheckBox()
            cb.setChecked(checked)
            cb.stateChanged.connect(top.show_or_hide_func(child.text(0)))
            top.tree.setItemWidget(child, 3, cb)
        for child in children:
            child.setSelected(True)
        for i, child in enumerate(map(top.child, range(last+1))):
            child.setText(1, u'⇧' if i > 0 else '')
            child.setText(2, u'⇩' if i < last else '')
        top.grouper.reorder(
            [child.text(0) for child in map(top.child,
                                            range(top.childCount()))])
        top.grouper.figure.canvas.draw()

    def add_tree(self, groupers, title=None):
        """Add a new QTreeWidget to the :attr:`tabs` widget"""
        tree = QTreeWidget(parent=self)
        tree.setColumnCount(7)
        tree.setHeaderLabels(
            ['Name', '', '', 'Visible', 'mean', 'min', 'max', 'Group'])
        ds = groupers[0].arrays[0].psy.base
        tree.itemClicked.connect(self.move_selected_children)
        tree.setSelectionMode(QAbstractItemView.MultiSelection)

        # fill the tree
        for grouper in groupers:
            top = GrouperItem(grouper, tree, 0)
            tree.addTopLevelItem(top)
            top.add_array_children()

        self.tabs.addTab(tree, title or 'Dataset %i' % ds.psy.num)
        tree.expandAll()
        for i in range(7):
            tree.resizeColumnToContents(i)
        self.show_plugin()
        self.groupers[ds.psy.num] = groupers
        self.trees[ds.psy.num] = tree



"""psy-strat: A psyplot plugin for stratigraphic plots

Describe your plugin here and do whatever you want. It should at least have a
__version__ attribute to specify the version of the package
"""

__version__ = "0.1.0"



"""plotters module of the psy-strat psyplot plugin

This module defines the plotters for the psy-strat package.
"""
from __future__ import division
import textwrap
from itertools import cycle
from psyplot.data import safe_list
from psyplot.plotter import (
    Formatoption, DictFormatoption, BEFOREPLOTTING, START)
import six
import psy_simple.plotters as psyps
from psy_simple.base import (
    TextBase, label_props, label_size, label_weight, Title)
import numpy as np

# -----------------------------------------------------------------------------
# ---------------------------- Formatoptions ----------------------------------
# -----------------------------------------------------------------------------


class LeftTitle(Title):
    """
    Show the title

    Set the title of the plot.
    %(replace_note)s

    Possible types
    --------------
    str
        The title for the :func:`~matplotlib.pyplot.title` function.

    Notes
    -----
    This is the title of this specific subplot! For the title of the whole
    figure, see the :attr:`figtitle` formatoption.

    See Also
    --------
    figtitle, titlesize, titleweight, titleprops"""
    # Reimplemented to plot the title on the left side

    def initialize_plot(self, value):
        arr = self.data
        self.texts = [self.ax.set_title(
            self.replace(value, arr, attrs=self.enhanced_attrs),
            loc='left')]


class TitleWrap(Formatoption):
    """
    Wrap the title automatically

    This formatoption wraps the title using :func:`textwrap.wrap`.

    Possible types
    --------------
    int
        If 0, the title will not be rapped, otherwise it will be wrapped after
        the given number of characters

    Notes
    -----
    This wraps the title after a certain amount of characters. For wrapping the
    text automatically before it leaves the plot area, use
    ``titleprops=dict(wrap=True)``
    """

    group = 'labels'

    name = 'Wrap the title'

    dependencies = ['title']

    def update(self, value):
        if value:
            for t in self.title.texts:
                t.set_text('\n'.join(textwrap.wrap(t.get_text(), value)))


class AxesGrouper(TextBase, Formatoption):
    """
    Group several axes through a bar

    This formatoption groups several plots on the same row by drawing a
    bar over them

    Possible types
    --------------
    None
        To not do anything
    tuple (float ``y``, str ``s``)
        A tuple of length 2, where the first parameter ``0<=y<=1`` determines
        the distance of the bar to the top y-axis and the second is the title
        of the group. `y` must be given relative to the axes height.
    """

    texts = []

    annotations = []

    value2share = None

    name = 'Group the axes'

    dependencies = ['titleprops', 'title']

    def initialize_plot(self, value):
        self.texts = []
        self.annotations = []
        super(AxesGrouper, self).initialize_plot(value)
        # redraw the annotation on resize to make sure it stays at the same
        # place as the text
        self.ax.figure.canvas.mpl_connect('resize_event', self.onresize)

    def update(self, value):
        self.remove()
        if value is None:
            return
        self.set_params(value)
        self.create_text(value)
        self.create_annotations()

    def create_text(self, value):
        if self.angle == 90:
            xstart = self.y_px
        else:
            xstart = self.y_px * np.tan(self.angle * np.pi / 180.)
        tr = self.ax.figure.transFigure.inverted()
        xstart = tr.transform((xstart, self.y_px))[0]
        self.texts.append(
            self.ax.text(xstart + (self.x0 + self.x1) / 2.0,
                         self.top + self.y,
                         self.replace(value[1], self.data),
                         ha='center', va='bottom',
                         transform=self.ax.figure.transFigure,
                         bbox=dict(facecolor='w', edgecolor='none')))

    def create_annotations(self):
        """Annotate from the left to the right axes"""
        fmtos = [self] + list(self.shared)
        boxes = [fmto.ax.get_position() for fmto in fmtos]
        fmto0 = min(zip(fmtos, boxes), key=lambda t: t[1].x0)[0]
        fmto1 = max(zip(fmtos, boxes), key=lambda t: t[1].x0)[0]
        t0 = fmto0.ax._left_title
        t1 = fmto1.ax._left_title
        kws = dict(
            zorder=self.texts[0].get_zorder() - 0.1,
            arrowprops=dict(
                arrowstyle="-",
                connectionstyle="angle,angleA=%1.3f,angleB=0" % self.angle))
        ax = self.ax
        self.annotations = [
            ax.annotate("", (0.0, 0.5), (0.0, 0.0), self.texts[0], t0, **kws),
            ax.annotate("", (1.0, 0.5), (0.0, 0.0), self.texts[0], t1, **kws)]

    def set_params(self, value):
        """Set the parameters for the annotation and the text"""
        y, s = value
        this_bbox = self.ax.get_position()
        if not self.shared:
            x0 = this_bbox.x0
            x1 = this_bbox.x1
            top = this_bbox.y1
        else:
            boxes = [this_bbox] + [
                fmto.ax.get_position() for fmto in self.shared]
            top = max(bbox.y1 for bbox in boxes)
            x0 = min(bbox.x0 for bbox in boxes)
            x1 = max(bbox.x0 for bbox in boxes)
        self.x0 = x0
        self.x1 = x1
        tr = self.ax.figure.transFigure
        self.y = y * this_bbox.height
        self.y_px = (tr.transform((x0, top + self.y))[1] -
                     tr.transform((x0, top))[1])
        self.top = top
        self.angle = self.titleprops.value.get('rotation', 45)

    def onresize(self, event):
        if self.shared:
            self.update(self.value)

    def remove(self, annotation=True, text=True):
        if text:
            for t in self.texts[:]:
                t.remove()
                self.texts.remove(t)
        if annotation:
            for a in self.annotations:
                a.remove()
            self.annotations.clear()


class AxisLineStyle(DictFormatoption):
    """
    Set the linestyle the x- and y-axes

    This formatoption sets the linestyle of the left, right, bottom and top
    axis.

    Possible types
    --------------
    dict
        Keys may be one of {'right', 'left', 'bottom', 'top'},  the values can
        be any valid linestyle or None to use the default style. The line style
        string can be one of (['solid' | 'dashed', 'dashdot', 'dotted'
        | (offset, on-off-dash-seq) | '-' | '--' | '-.' | ':' | 'None' | ' ' |
        ''])."""

    group = 'axes'

    name = 'Linestyle of x- and y-axes'

    @property
    def value2pickle(self):
        """Return the current axis colors"""
        return {key: s.get_linestyle() for key, s in self.ax.spines.items()}

    def initialize_plot(self, value):
        positions = ['right', 'left', 'bottom', 'top']
        #: :class:`dict` storing the default linewidths
        self.default_lw = dict(zip(positions, map(
            lambda pos: self.ax.spines[pos].get_linewidth(), positions)))
        self.update(value)

    def update(self, value):
        for pos, style in six.iteritems(value):
            spine = self.ax.spines[pos]
            spine.set_linestyle(style)
            if self is not None and spine.get_linewidth() == 0.0:
                spine.set_linewidth(1.0)
            elif self is None:
                spine.set_linestyle('solid')
                spine.set_linewidth(self.default_lw[pos])


class MeasurementLines(Formatoption):
    """
    Draw lines at the measurement locations

    Possible types
    --------------
    None
        Don't draw any lines
    color
        The color of the lines
    """

    default = None

    artists = None

    dependencies = ['transpose', 'xlim', 'plot']

    def update(self, value):
        self.remove()
        if value is None:
            return
        get_y = self.transpose.get_y
        ys = np.unique(np.concatenate([get_y(arr) for arr in self.iter_data]))
        if self.plot.value is not None:
            kws = {'zorder': self.plot._plot[0].get_zorder() - 0.2}
        else:
            kws = {}
        self.artists = self.ax.hlines(ys, *self.xlim.range, color=value, **kws)

    def remove(self):
        if self.artists is not None:
            self.artists.remove()
            self.artists = None


class ExagFactor(Formatoption):
    """
    The exaggerations factor

    Possible types
    --------------
    float
        The factor by how much the data should be exaggerated

    See Also
    --------
    exag_color, exag
    """

    priority = BEFOREPLOTTING

    name = 'Exaggeration factor'

    def update(self, value):
        # Does nothing
        pass


class ExagPlot(psyps.LinePlot):

    __doc__ = psyps.LinePlot.__doc__

    dependencies = ['exag_factor']

    def plot_arr(self, arr, *args, **kwargs):
        return super(ExagPlot, self).plot_arr(
            arr * self.exag_factor.value, *args, **kwargs)


class Occurences(Formatoption):
    """
    Specify the range for occurences

    This formatoption can be used to specify a minimum and a maximum value.
    The parts of the data that fall into this range will be considered as an
    occurence, set to 0 and marked by the :attr:`occurence_value` formatoption

    Possible types
    --------------
    None
        Do not mark anything as an occurence
    float
        Anything below the given number will be considered as an occurence
    tuple of floats ``(vmin, vmax)``
        The minimum and maximum value. Anything between `vmin` and `vmax` will
        be marked as a occurence

    See Also
    --------
    occurence_marker, occurence_value
    """

    name = 'Occurences range'

    priority = START

    children = ['maskless', 'maskleq', 'maskgreater', 'maskgeq']

    def update(self, value):
        if value is None:
            return
        self.occurences = []
        try:
            value = list(value)
        except TypeError:
            value = [-np.inf, value]
        vmin, vmax = value
        for i, arr in enumerate(self.iter_data):
            new_arr = arr.copy(True)
            mask = (arr.values >= vmin) & (arr.values <= vmax)
            self.occurences.append(mask)
            new_arr.values[mask] = 0
            self.set_data(new_arr, i)


class OccurenceMarker(Formatoption):
    """
    Specify the marker for occurences

    This formatoption can be used to define the marker style for occurences.

    Possible types
    --------------
    None
        Use the mean of the axes limits
    float
        Specify the x-value for an occurence
    list of floats
        Specify the x-value for an occurence for each array explicitly

    See Also
    --------
    occurences, occurence_value
    """

    name = 'Marker for the occurences'

    def update(self, value):
        # Does nothing, value is used in :meth:`OccurencePlot.update`
        pass


class OccurencePlot(Formatoption):
    """
    Specify the value to use for occurences in the plot

    This formatoption can be used to define where the occurence marker should
    be placed.

    Possible types
    --------------
    None
        Use the mean of the axes limits
    float
        Specify the x-value for an occurence
    list of floats
        Specify the x-value for an occurence for each array explicitly

    See Also
    --------
    occurences, occurence_marker
    """

    dependencies = ['occurences', 'xlim', 'occurence_marker', 'color',
                    'transpose']

    _artists = None

    name = 'Occurence plot value'

    def update(self, value):
        self.remove()
        if self.occurences.value is None:
            return
        elif value is None:
            value = np.mean(self.xlim.range)
        self._artists = artists = []
        for mask, arr, val, color, marker in zip(
                self.occurences.occurences, self.iter_data,
                cycle(safe_list(value)), self.color.colors,
                cycle(self.occurence_marker.value)):
            x = arr[arr.dims[-1]].values[mask]
            y = [val] * len(x)
            if self.transpose.value:
                x, y = y, x
            artists.extend(
                self.ax.plot(x, y, marker=marker, color=color, lw=0))

    def remove(self):
        if self._artists is not None:
            for a in self._artists:
                try:
                    a.remove()
                except ValueError:
                    pass
            del self._artists


# -----------------------------------------------------------------------------
# ------------------------------ Plotters -------------------------------------
# -----------------------------------------------------------------------------


class StratPlotter(psyps.LinePlotter):
    """A plotter for stratigraphic diagrams"""

    _rcparams_string = ['plotter.strat.']

    axislinestyle = AxisLineStyle('axislinestyle')
    title = LeftTitle('title')
    title_wrap = TitleWrap('title_wrap')
    grouper = AxesGrouper('grouper')
    grouperprops = label_props(grouper)
    grouperweight = label_weight(grouper)
    groupersize = label_size(grouper)
    hlines = MeasurementLines('hlines')
    exag_color = psyps.LineColors('exag_color')
    exag_factor = ExagFactor('exag_factor')
    exag = ExagPlot('exag', color='exag_color')

    # occurences
    occurences = Occurences('occurences')
    occurence_marker = OccurenceMarker('occurence_marker')
    occurence_value = OccurencePlot('occurence_value')


class BarStratPlotter(psyps.BarPlotter):
    """A bar plotter for stratigraphic diagrams"""

    _rcparams_string = ['plotter.strat.', 'plotter.barstrat.']

    axislinestyle = AxisLineStyle('axislinestyle')
    title = LeftTitle('title')
    title_wrap = TitleWrap('title_wrap')
    grouper = AxesGrouper('grouper')
    grouperprops = label_props(grouper)
    grouperweight = label_weight(grouper)
    groupersize = label_size(grouper)
    hlines = MeasurementLines('hlines')
    exag_color = psyps.LineColors('exag_color')
    exag_factor = ExagFactor('exag_factor')
    exag = ExagPlot('exag', color='exag_color')

    # occurences
    occurences = Occurences('occurences')
    occurence_marker = OccurenceMarker('occurence_marker')
    occurence_value = OccurencePlot('occurence_value')



"""Module to create new stratigraphic plots

This module defines the :func:`stratplot` function that can be used to create
stratigraphic plots such as pollen diagrams
"""
from __future__ import division
import weakref
import six
from itertools import groupby, chain, islice
import matplotlib as mpl
import matplotlib.transforms as mt
from collections import defaultdict
import psyplot
from psyplot.utils import DefaultOrderedDict
import xarray as xr
import numpy as np
from psy_strat.plotters import StratPlotter, BarStratPlotter
from psyplot.data import ArrayList
import psyplot.project as psy
from docrep import DocstringProcessor

docstrings = DocstringProcessor()


gui_plugin = 'psy_strat.strat_widget:StratPlotsWidget:stratplots'


NOGROUP = 'nogroup'


def _no_grouper(col):
    """Return an empty string to disable the grouping"""
    return NOGROUP


def stratplot(df, group_func=None, formatoptions=None, ax=None,
              thresh=0.01, percentages=[], exclude=[],
              widths=None, calculate_percentages=True,
              min_percentage=20.0, trunc_height=0.3, fig=None, all_in_one=[],
              stacked=[], summed=[], use_bars=False, subgroups={}):
    """Visualize a dataframe as a stratigraphic plot

    This functions takes a :class:`pandas.DataFrame` and transforms it to a
    stratigraphic plot. The columns in the DataFrame may be grouped together
    using the `group_func` and the widths per group should then be specified.
    This function uses matplotlib axes for each subdiagram that all share a
    common vertical axes, the index of `df`. The variables are managed in the
    order of occurence in the input `df` but, however, are grouped together
    depending on the `group_func`.

    The default is to plot every variable in `df` into separete line plots
    that line up vertically. You can use the `percentages` parameter for area
    plots, the `all_in_one` parameter for groups that should be all in one
    single plot (i.e. axes) and the `stacked` parameter for stacked plots.

    Parameters
    ----------
    df: pandas.DataFrame
        The dataframe containing the data to plot.
    group_func: function
        A function that groups the columns in the input `df` together. It must
        accept the name of a column and return the corresponding group name::

            def group_func(col_name: str):
                return "name of it's group"

        If this parameter is not specified, each column will be assigned to the
        `'nogroup'` group that can then be used in the other parameters, such
        as `formatoptions` and `percentages`. Each group may also be divided
        into `subgroups` (see below), in this case, the `group_func` should
        return the corresponding subgroup.
    formatoptions: dict
        The formatoption for each group. Depending on the chosen plot method,
        this contains the formatoptions for the psyplot plotter.
    ax: matplotlib.axes.Axes
        The matplotlib axes to plot on. New axes will be created that cover all
        the space of the given axes.
        If this parameter is not specified and `fig` is None, a new matplotlib
        figure is created with a new matplotlib axes.
    thresh: float
        A minimum number between 0 and 100 (by default 1%) that a
        `percentages` column has to fullfil in order to be included in the
        plot. If a variable is always below this threshold, it will not be
        included in the figure
    percentages: list of str
        The group names (see `group_func`) that represent percentage values.
        This variables will be visualized using an area plot and can be
        rescaled to sum up to 100% using the `calculate_percentages` parameter.
    exclude: list of str
        Either group names of column names in `df` that should be excluded in
        the plot
    widths: dict
        A mapping from group name to it's relative width in the plot. The
        values of this mapping should some up to 1, e.g.::

            widths = {'group1': 0.3, 'group2': 0.5, 'group3': 0.2}
    calculate_percentages: bool or list of str
        If True, rescale the groups mentioned in the `percentages` parameter
        to sum up to 100%. In case of a list of str, this parameter represents
        the group (or variable) names that shall be used for the normalization
    min_percentage: float
        The minimum percentage (between 0 and 100) that should be covered by
        variables displaying `percentages` data. Each plot in one of the
        `percentages` groups will have at least have a xlim from 0 to
        `min_percentage`
    trunc_height: float
        A float between 0 and 1. The fraction of the `ax` that should be
        reserved for the group titles.
    fig: matplotlib.Figure
        The matplotlib figure to draw the plot on. If neither `ax` nor `fig` is
        specified, a new figure will be created.
    all_in_one: list of str
        The groups mentioned in this parameter will all be plotted in one
        single axes whereas the default is to plot each variable in a separate
        plot
    stacked: list of str
        The groups mentioned in this parameter will all be plotted in one
        single axes, stacked onto each other
    summed: list of str
        The groups (or subgroups) mentioned in this parameter will be summed
        and an extra plot will be appended to the right of the stratigraphic
        diagram
    use_bars: list of str or bool
        The variables specified in this parameter (or all variables if
        `use_bars` is ``True``) will be visualized by a bar diagram, instead
        of a line or area plot.
    subgroups: dict
        A mapping from group name to a list of subgroups, e.g.::

            subgroups = {'Pollen': ['Trees', 'Shrubs']}

        to divide an overarching group into subgroups.

    Returns
    -------
    psyplot.project.Project
        The newly created psyplot subproject that contains the displayed data
    list of :class:`StratGroup`
        The groupers that manage the different variables. There is one
        grouper per group"""
    import psyplot.project as psy
    import matplotlib.pyplot as plt
    if group_func is None:
        group_func = _no_grouper
    groups = DefaultOrderedDict(list)
    # we invert subgroups here
    subgroup2group = dict(chain.from_iterable(
        ((sub, group) for sub in subs)
        for group, subs in subgroups.items()))
    cols = {}
    for col in df.columns:
        group = group_func(col)
        group = subgroup2group.get(group, group)
        groups[group].append(col)
        cols[col] = group

    formatoptions = formatoptions or {}
    if calculate_percentages and set(percentages).intersection(groups):
        df = df.copy(True)
        for group in set(percentages).intersection(groups):
            members = groups[group]
            try:
                calculate_percentages = list(calculate_percentages)
            except TypeError:
                norm_members = members
            else:
                norm_members = list(set(chain.from_iterable(
                    [var] if var in df.columns else groups[var]
                    for var in calculate_percentages)))

            df[members] *= 100. / np.tile(
                df[norm_members].fillna(0).sum(axis=1)[:, np.newaxis],
                (1, len(members)))
    if summed:
        try:
            summed = list(summed)
        except TypeError:
            summed = list(groups)
        stacked.append('Summed')
        groups['Summed'] = [g + '_summed' for g in summed]
        if widths:
            widths.setdefault('Summed', 0.2)
        formatoptions.setdefault('Summed', {}).setdefault(
            'legendlabels', '%(long_name)s')
        formatoptions['Summed'].setdefault('title', '')
    else:
        summed = []

    widths = widths or defaultdict(
        lambda: 1. / (len(set(groups).difference(percentages)) or 1))

    # Setup use_bars
    if isinstance(use_bars, six.string_types):
        use_bars = [use_bars]
    try:
        use_bars = list(use_bars)
    except TypeError:
        use_bars = list(groups) if use_bars else []

    # NOTE: we create the Dataset manually instead of using
    # xarray.Dataset.from_dataframe becuase that is much faster
    idx = df.index.name or 'y'
    ds = xr.Dataset(
        {col: xr.Variable((idx, ), df[col]) for col in df.columns},
        {idx: xr.Variable((idx, ), df.index)})
    for var, varo in ds.variables.items():
        if var not in ds.coords:
            varo.attrs['group'] = group_func(var)
            varo.attrs['maingroup'] = cols[var]
    for group in summed:
        variables = [var for var, varo in ds.variables.items()
                     if varo.attrs.get('group') == group]
        ds[group + '_summed'] = xr.Variable(
            (idx, ), df[variables].sum(axis=1).values,
            attrs={'long_name': group, 'group': 'Summed',
                   'maingroup': 'Summed'})

        cols[group + '_summed'] = 'Summed'

    plot_vars = [
        var for var, varo in ds.variables.items()
        if ((var not in ds.coords) and
            (var not in exclude and varo.attrs['group'] not in exclude) and
            (cols[var] not in percentages or ds[var].max().values > thresh))]
    arr_names = []

    if ax is None:
        fig = fig or plt.figure()
        bbox = mt.Bbox.from_extents(
            mpl.rcParams['figure.subplot.left'],
            mpl.rcParams['figure.subplot.bottom'],
            mpl.rcParams['figure.subplot.right'],
            mpl.rcParams['figure.subplot.top'])
    elif isinstance(ax, (mpl.axes.SubplotBase, mpl.axes.Axes)):
        bbox = ax.get_position()
        fig = ax.figure
    else:  # the bbox is given
        bbox = ax
        fig = fig or plt.gcf()
    x0 = bbox.x0
    y0 = bbox.y0
    orig_height = bbox.height
    height = orig_height * (1 - trunc_height)
    total_width = bbox.width
    x1 = x0 + total_width

    i = 0
    ax0 = None
    x = x0
    mp = psy.gcp(True)
    groupers = []
    with psy.Project.block_signals:
        for group, variables in groups.items():

            variables = [v for v in variables if v in plot_vars]
            if not variables:
                continue
            w = widths[group] * total_width
            if group in all_in_one:
                identifier = 'all_in_one'
            elif group in stacked:
                identifier = 'stacked'
            elif group in percentages:
                identifier = 'percentages'
            else:
                identifier = 'default'
            grouper_cls = strat_groupers[identifier]
            grouper = grouper_cls.from_dataset(
                fig, mt.Bbox.from_bounds(x, y0, w, height),
                ds, variables, fmt=dict(formatoptions.get(group, {})),
                project=mp, ax0=ax0, use_bars=use_bars, group=group)
            if identifier == 'percentages':
                resize = False
                for plotter in grouper.plotters:
                    if plotter.ax.get_xlim()[1] < min_percentage:
                        plotter.update(xlim=(0, min_percentage))
                        resize = True
                if resize:
                    grouper.resize_axes(grouper.axes)
            if group != NOGROUP:
                grouper.group_plots(trunc_height / height)
            ds[group] = xr.Variable(tuple(), '',
                                    attrs={'identifier': identifier})
            ax0 = ax0 or grouper.axes[0]
            x += w

            arr_names.extend(
                arr.psy.arr_name for arr in grouper.plotter_arrays)
            groupers.append(grouper)
        if psyplot.with_gui:
            from psyplot_gui.main import mainwindow
            mainwindow.plugins[gui_plugin].add_tree(groupers)
    # invert the vertical axis
    ax0.invert_yaxis()

    sp = psy.gcp(True)(arr_name=arr_names)
    sp[0].psy.update(
        ylabel='%(name)s', ytickprops={'left': True, 'labelleft': True},
        draw=False)
    for ax, p in sp.axes.items():
        ax_bbox = ax.get_position()
        d = {}
        if ax_bbox.x0 != x0:
            d['left'] = ':'
        if ax_bbox.x1 != x1:
            d['right'] = ':'
        p.update(axislinestyle=d, draw=False)
    psy.scp(sp.main)
    psy.scp(sp)
    return sp, groupers


class StratGroup(object):
    """Base class for visualizing stratigraphic plots"""

    #: list of weakref. Weak references to the created arrays
    _refs = []

    _arrays = None

    _plotter_arrays = None

    grouper_height = None

    #: The default formatoptions for the plots
    default_fmt = {
        'ytickprops': {'left': False, 'labelleft': False},
        }

    bar_default_fmt = default_fmt.copy()
    bar_default_fmt['categorical'] = False

    @property
    def plotter_arrays(self):
        """The data objects that contain the plotters"""
        return self._plotter_arrays or ArrayList([ref() for ref in self._refs])

    @plotter_arrays.setter
    def plotter_arrays(self, value):
        self._plotter_arrays = value

    @property
    def arrays(self):
        """The arrays managed by this :class:`StratGroup`. One array for each
        variable"""
        return self.plotter_arrays

    @property
    def all_arrays(self):
        """All variables of this group in the dataset"""
        arr = self.arrays[0]
        group = arr.group
        ds = arr.psy.base
        return [ds.psy[arr] for arr, v in ds.variables.items()
                if v.attrs.get('group') == group]

    @property
    def plotters(self):
        """The plotters of the :attr:`arrays`"""
        return list(filter(lambda p: p is not None,
                           [arr.psy.plotter for arr in self.plotter_arrays]))

    @property
    def axes(self):
        return [plotter.ax for plotter in self.plotters]

    @property
    def arr_names(self):
        return [arr.psy.arr_name for arr in self.plotter_arrays]

    def __init__(self, arrays, bbox=None, use_weakref=True, group=None):
        """
        Parameters
        ----------
        arrays: list of xarray.DataArray
            The data arrays that are plotted by this :class:`StratGroup`
            instance
        bbox: matplotlib.transforms.Bbox
            The bounding box for the axes
        use_weakref: bool
            If True, only weak references are used
        group: str
            The groupname of this grouper. If not given, it will be taken from
            the ``'maingroup'`` attribute of the first array
        """
        if use_weakref:
            self._refs = [weakref.ref(arr) for arr in arrays]
        else:
            self.plotter_arrays = arrays
        if bbox is None:
            boxes = [arr.psy.ax.get_position() for arr in arrays]
            x0 = min(bbox.x0 for bbox in boxes)
            y0 = min(bbox.y0 for bbox in boxes)
            w = sum(bbox.width for bbox in boxes)
            bbox = boxes[0].from_bounds(x0, y0, w, boxes[0].height)
        self.bbox = bbox
        self.group = group or arrays[0].attrs.get('maingroup')

    def resize_axes(self, axes):
        """Resize the axes in this group"""
        width = self.bbox.width
        w = width / len(axes)
        x0 = self.bbox.x0
        for ax in axes:
            ax_bbox = ax.get_position()
            ax.set_position([x0, ax_bbox.y0, w, ax_bbox.height])
            x0 += w

    def group_plots(self, height=None):
        """Group the variables visually

        Parameters
        ----------
        height: float
            The height of the grouper. If not specified, the previous
            :attr:`grouper_height` attribute will be used"""
        for plotter in (plotter for plotter in self.plotters
                        if plotter.ax.get_visible() and
                        not plotter.grouper.shared_by):
            height = height or self.grouper_height
            if height is None and plotter['grouper']:
                height = plotter['grouper'][0]
            elif height is None:
                return
            self.grouper_height = height
            plotter.update(grouper=(height, '%(group)s'),
                           draw=False, force=True)
            for fmto in plotter.grouper.shared:
                with fmto.plotter.no_validation:
                    fmto.plotter['grouper'] = (height, '%(group)s')

    @property
    def figure(self):
        """The figure that contains the plots"""
        return self.axes[0].figure

    def is_visible(self, arr):
        """Check if the given `arr` is shown"""
        return arr.psy.plotter.ax.get_visible()

    @classmethod
    @docstrings.get_sectionsf('StratGroup.from_dataset',
                              sections=['Parameters', 'Returns'])
    def from_dataset(cls, fig, bbox, ds, variables, fmt=None, project=None,
                     ax0=None, use_bars=False, group=None):
        """
        Create :class:`StratGroup` while creating a stratigraphic plot

        Create a stratigraphic plot within the given `bbox` of `fig`.

        Parameters
        ----------
        fig: matplotlib.figure.Figure
            The figure to plot in
        bbox: matplotlib.transforms.Bbox
            The bounding box for the newly created axes
        ds: xarray.Dataset
            The dataset
        variables: list
            The variables that shall be plot in the given `ds`
        project: psyplot.project.Project
            The mother project. If given, only weak references are stored in
            the returned :class:`StratGroup` and each array is appended to the
            `project`.
        ax0: matplotlib.axes.Axes
            The first subplot to share the y-axis with
        use_bars: bool
            Whether to use a bar plot or a line/area plot

        Returns
        -------
        StratGroup
            The newly created instance with the arrays
        """
        if ax0 is None:
            ax0 = fig.add_axes(bbox.from_bounds(*bbox.bounds), label='ax0')
            axes = [ax0] + [fig.add_axes(bbox.from_bounds(*bbox.bounds),
                                         sharey=ax0, label='ax%i' % i)
                            for i in range(1, len(variables))]
        else:
            axes = [fig.add_axes(bbox.from_bounds(*bbox.bounds),
                                 sharey=ax0, label='ax%i' % i)
                    for i in range(len(variables))]
        grouped = DefaultOrderedDict(list)
        for name in variables:
            grouped[ds[name].attrs.get('group', 'group')].append(name)
        # Use group specific bars
        if use_bars:
            try:
                use_bars = list(use_bars)
            except TypeError:
                use_bars = list(grouped)
        else:
            use_bars = []
        sp = None
        axes_it = iter(axes)
        for subgroup, names in grouped.items():
            formatoptions = dict(fmt or {})
            if subgroup in use_bars or group in use_bars:
                plotter_cls = BarStratPlotter
                defaults = cls.bar_default_fmt
            else:
                plotter_cls = StratPlotter
                defaults = cls.default_fmt
            for key, val in six.iteritems(defaults):
                formatoptions.setdefault(key, val)
            sp2 = psy.Project()._add_data(
                plotter_cls, ds, name=names, draw=False, fmt=formatoptions,
                prefer_list=False, ax=islice(axes_it, len(names)),
                share='grouper', attrs=dict(maingroup=group))
            if project is not None:
                project.extend(sp2, new_name=True)
            sp = sp2 if sp is None else sp + sp2
        ret = cls(list(sp), bbox, use_weakref=project is not None,
                  group=group)
        ret.resize_axes(axes)
        return ret

    def hide_array(self, name):
        """Hide the variable of the given `name`

        Parameters
        ----------
        name: str
            The variable name"""
        arr = next(iter(self.plotter_arrays(name=name)), None)
        if arr is None:
            return
        group = arr.group
        i, first_visible = next(filter(
            lambda t: t[1].psy.ax.get_visible() and t[1].group == group,
            enumerate(self.plotter_arrays)))
        if arr is None or not arr.psy.ax.get_visible():  # array isn't plotted
            return
        elif arr is first_visible:
            p = psy.Project(self.plotter_arrays)(group=group)
            p.unshare(keys='grouper', draw=False)
            p(name=set(p.names) - {name}).share(keys='grouper', draw=False)
        arr.psy.ax.set_visible(False)
        if arr is self.arrays[0]:
            pass
        self.resize_axes([ax for ax in self.axes if ax.get_visible()])
        self.group_plots()

    def show_array(self, name):
        """Show the variable of the given `name`

        Parameters
        ----------
        name: str
            The variable name"""
        arrays = self.plotter_arrays
        arr = next(iter(arrays(name=name)), None)
        if arr is None:
            return
        group = arr.group
        key, first_invisibles = next(
            groupby(filter(lambda a: a.group == group, arrays),
                    lambda arr: arr.psy.ax.get_visible()))
        if key:  # first plot is visible
            first_invisibles = []

        if arr.psy.ax.get_visible():  # array isn't plotted
            return
        elif any(arr is invisible_arr for invisible_arr in first_invisibles):
            p = psy.Project(arrays)(group=group)
            p.unshare(keys='grouper', draw=False)
#            i = next(i for i, a in enumerate(p) if a.name == name)
            p(name=set(p.names) - {name}).share(arr, keys='grouper',
                                                draw=False)
        arr.psy.ax.set_visible(True)
        self.resize_axes([ax for ax in self.axes if ax.get_visible()])
        self.group_plots()

    def reorder(self, names):
        """Reorder the plot objects

        Parameters
        ----------
        names: list of str
            The variable names that should be the first"""
        arrays = self._plotter_arrays or self._refs
        old = list(arrays)
        old_da = list(self.plotter_arrays)
        project = self.plotters[0].project
        if project is not None:
            project = project.main
            i = project.arr_names.index(old_da[0].psy.arr_name)
            arr_names = project.arr_names[i:i+len(arrays)]
            reorder_project = arr_names == self.arr_names
        arrays.clear()
        for name in names:
            arr = next((old[i] for i, arr in enumerate(old_da)
                        if str(arr.name) == name), None)
            if arr is not None:
                arrays.append(arr)
        # now add the ones that are not mentioned in `names`
        for i, arr in enumerate(old_da):
            if str(arr.name) not in names:
                arrays.append(old[i])
        if project is not None and reorder_project:
            project[i:i+len(arrays)] = self.plotter_arrays
            if project.is_csp or project.is_cmp:
                project.oncpchange.emit(project)
        self.resize_axes([ax for ax in self.axes if ax.get_visible()])
        self.plotter_arrays.update(force=['grouper'], draw=False)


class StratPercentages(StratGroup):
    """A :class:`StratGroup` for percentages plots"""

    default_fmt = StratGroup.default_fmt.copy()
    default_fmt['xlim'] = (0, 'rounded')
    default_fmt['xticks'] = np.arange(10, 100, 20)

    bar_default_fmt = default_fmt.copy()
    bar_default_fmt['categorical'] = False

    default_fmt['plot'] = 'areax'

    def resize_axes(self, axes):
        """Resize the axes in this group"""
        width = self.bbox.width
        width /= sum(ax.get_xlim()[1] for ax in axes) / 100.
        x0 = self.bbox.x0
        for ax in axes:
            w = width * ax.get_xlim()[1] / 100.
            ax_bbox = ax.get_position()
            ax.set_position([x0, ax_bbox.y0, w, ax_bbox.height])
            x0 += w


class StratAllInOne(StratGroup):
    """A :class:`StratGroup` for single plots"""

    default_fmt = StratGroup.default_fmt.copy()
    default_fmt['title'] = '%(group)s'
    default_fmt['titleprops'] = {}
    default_fmt['legend'] = True

    bar_default_fmt = default_fmt.copy()
    bar_default_fmt['categorical'] = False

    @property
    def arrays(self):
        return self.plotter_arrays[0]

    def group_plots(self, height):
        """Reimplemented to do nothing because all variables are in one axes
        """
        pass

    def is_visible(self, arr):
        """Check if the given `arr` is shown"""
        return arr.name in self.plotter_arrays[0].names

    @classmethod
    @docstrings.dedent
    def from_dataset(cls, fig, bbox, ds, variables, fmt=None, project=None,
                     ax0=None, use_bars=False, group=None):
        """
        Create :class:`StratGroup` while creating a stratigraphic plot

        Create a stratigraphic plot within the given `bbox` of `fig`.

        Parameters
        ----------
        %(StratGroup.from_dataset.parameters)s

        Returns
        -------
        %(StratGroup.from_dataset.returns)s
        """
        fmt = fmt or {}
        if use_bars:
            try:
                use_bars = list(iter(use_bars))
            except TypeError:
                use_bars = True
            else:
                use_bars = group in use_bars
        defaults = cls.bar_default_fmt if use_bars else cls.default_fmt
        for key, val in six.iteritems(defaults):
            fmt.setdefault(key, val)
        if ax0 is None:
            ax = fig.add_axes(bbox.from_bounds(*bbox.bounds), label='ax0')
        else:
            ax = fig.add_axes(bbox.from_bounds(*bbox.bounds),
                              sharey=ax0, label='ax0')
        plotter_cls = BarStratPlotter if use_bars else StratPlotter
        sp = psy.Project()._add_data(
            plotter_cls, ds, name=variables, draw=False, fmt=fmt,
            prefer_list=True, ax=ax, share='grouper',
            attrs=dict(maingroup=group))
        if project is not None:
            project.extend(sp, new_name=True)
        return cls(list(sp), bbox, use_weakref=project is not None,
                   group=group)

    def hide_array(self, name):
        """Hide the variable of the given `name`

        Parameters
        ----------
        name: str
            The variable name"""
        i, arr = next(((i, arr) for i, arr in enumerate(self.arrays)
                       if arr.name == name), (None, None))
        plotter = self.plotters[0]
        v = plotter['plot']
        if v is None or isinstance(v, six.string_types):
            v = [v] * len(self.arrays)
        if arr is None or v[i] is None:  # array isn't plotted
            return
        v[i] = None
        plotter.update(plot=v, force=True)

    def show_array(self, name):
        """Show the variable of the given `name`

        Parameters
        ----------
        name: str
            The variable name"""
        i, arr = next(((i, arr) for i, arr in enumerate(self.arrays)
                       if arr.name == name), (None, None))
        plotter = self.plotters[0]
        v = plotter['plot']
        if v is None or isinstance(v, six.string_types):
            v = [v] * len(self.arrays)
        if arr is None or v[i] is not None:  # array is plotted
            return
        v[i] = self.default_fmt.get('plot', '-')
        plotter.update(plot=v, force=True)

    def reorder(self, names):
        """Reorder the plot objects

        Parameters
        ----------
        mapping: dict
            A mapping from the new index to the old one"""
        plotter = self.plotters[0]
        data = plotter.data
        current = list(data)
        visibilities = list(map(self.is_visible, data))
        plot = []
        data.clear()
        ls = self.default_fmt.get('plot', '-')
        for name in names:
            i = next(i for i, arr in enumerate(current)
                     if str(arr.name) == name)
            data.append(current[i])
            plot.append((visibilities[i] and ls) or None)
        plotter.update(plot=plot, replot=True, draw=False)


class StackedGroup(StratAllInOne):
    """A grouper for stacked plots"""

    default_fmt = StratAllInOne.default_fmt.copy()
    default_fmt['plot'] = 'stacked'

    bar_default_fmt = StratAllInOne.bar_default_fmt.copy()
    bar_default_fmt['plot'] = 'stacked'


strat_groupers = {
    'all_in_one': StratAllInOne,
    'percentages': StratPercentages,
    'default': StratGroup,
    'stacked': StackedGroup}



"""Setup file for plugin psy-strat

This file is used to install the package to your python distribution.
Installation goes simply via::

    python setup.py install
"""

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys


def readme():
    with open('README.rst') as f:
        return f.read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


setup(name='psy-strat',
      version='0.1.0',
      description='A psyplot plugin for stratigraphic plots',
      long_description=readme(),
      classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
      ],
      keywords='visualization psyplot pollen',
      url='https://github.com/Chilipp/psy-strat',
      author='Philipp Sommer',
      author_email='philipp.sommer@unil.ch',
      license="GPLv2",
      packages=find_packages(exclude=['docs', 'tests*', 'examples']),
      install_requires=[
          'psyplot>=1.2.0',
          'psy-simple>=1.2.0',
      ],
      tests_require=['pytest'],
      cmdclass={'test': PyTest},
      entry_points={
          'psyplot': ['plugin=psy_strat.plugin'],
          'psyplot_gui': [
              'stratplots=psy_strat.strat_widget:StratPlotsWidget'],
          },
      zip_safe=False)
